import pytest
import pandas as pd
from chaindl.scraper.chainexposed import _download
from chaindl.scraper.chainexposed import _extract_data_from_scripts
from chaindl.scraper import utils

def test_extract_data_from_scripts():
    scripts = [
        type('', (object,), {"string": """Plotly.newPlotvar trace0 =
            {
            x: ["2012-01-01","2012-01-02","2012-01-03"],
            y: ["912.0","259.0","1375.0"],
            name: 'Price',
        }
                             
        Plotly.newPlot(target_target, data, layout, config); """})()
    ]

    expected_data = pd.DataFrame({
        "Price": [912.0, 259.0, 1375.0]
    }, index=pd.to_datetime(["2012-01-01","2012-01-02","2012-01-03"]))
    expected_data.index.name = 'Date'

    dfs = _extract_data_from_scripts(scripts)

    pd.testing.assert_frame_equal(dfs[0], expected_data)

def test_chainexposed_mocked(monkeypatch):
    def mock_get_page_content(url):
        return """
        <script>
        Plotly.newPlotvar trace0 =
            {
            x: ["2012-01-01","2012-01-02","2012-01-03"],
            y: ["912.0","259.0","1375.0"],
            name: 'Price',
        }
                             
        Plotly.newPlot(target_target, data, layout, config); 
        </script>
        """
    
    monkeypatch.setattr(utils, "_get_page_content", mock_get_page_content)

    url = "https://chainexposed.com/test"
    data = _download(url)

    expected_data = pd.DataFrame({
        "Price": [912.0, 259.0, 1375.0]
    }, index=pd.to_datetime(["2012-01-01","2012-01-02","2012-01-03"]))
    expected_data.index.name = 'Date'

    pd.testing.assert_frame_equal(data, expected_data)

@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://chainexposed.com/ConsolidatedMetricCoinsDestroyed.html",
        ['Price', 'Coins Destroyed', 'Coins Destroyed 7d MA']
    ),
    (
        "https://chainexposed.com/addressBandTotHUMPBACK.html",
        ['Price', '10000 and more (Humpbacks)']
    )
])
def test_real_chainexposed_download(url, expected_columns):
    data = _download(url)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert all(data.dtypes == float)

    assert all(col in data.columns for col in expected_columns)



import os
import pytest
import pandas as pd
from dotenv import load_dotenv
from chaindl.scraper.bitbo import _get_traces, _get_data, _download

load_dotenv()

sbr_webdriver = os.getenv('SBR_WEBDRIVER')

content = """
    var axis_x = ["2020-10-20","2020-10-21","2020-10-22"];
    var price_c = [11921.78,12813.11,12990.25];
    var sma200 = [9844.96,9874.65,9905.69];
    var sma1458 = [6650.14,6658.48,6666.94];
    var current = "2024-10-01";

    var trace1 = {
			type: "scatter",
			name: 'Price end of day',
			x: axis_x,
			y: price_c,
			line: {width: 2}
		};
    var trace2 = {
			type: "scatter",
			name: 'MA1458d',
			x: axis_x,
			y: sma1458,
			line: {width: 2}
		};
"""

def test_get_traces():
    traces = _get_traces(content)

    assert len(traces) == 2
    assert 'Price end of day' in traces[0]
    assert 'MA1458d' in traces[1]

def test_get_data():
    trace = '{\n\t\t\ttype: "scatter",\n\t\t\tname: \'Price end of day\',\n\t\t\tx: axis_x,\n\t\t\ty: price_c,\n\t\t\tline: {width: 2}\n\t\t}'
    
    name, x, y = _get_data(trace, content)

    assert name == 'Price end of day'
    assert x == ["2020-10-20","2020-10-21","2020-10-22"]
    assert y == [11921.78, 12813.11, 12990.25]

def test_download_data(monkeypatch):
    expected_df = pd.DataFrame({
        'Price end of day': [11921.78, 12813.11, 12990.25],
        'MA1458d': [6650.14, 6658.48, 6666.94]
    }, index=pd.to_datetime(["2020-10-20","2020-10-21","2020-10-22"]))
    expected_df.index.name = 'Date'

    import chaindl.scraper.bitbo as bitbo
    monkeypatch.setattr(bitbo, '_get_script_content', lambda url: content)

    result_df = _download('test_url')

    pd.testing.assert_frame_equal(result_df, expected_df)


@pytest.mark.skipif(not sbr_webdriver, reason="Skipping Bitbo integration tests: SBR_WEBDRIVER not provided.")
@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://charts.bitbo.io/cycle-repeat/",
        ['MA1458d', 'MA200d', 'Price end of day']
    ),
    (
        "https://charts.bitbo.io/macd/",
        ['Price', 'MACD', 'Signal line']
    )
])
def test_bitbo_download_sb(url, expected_columns):
    data = _download(url)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert all(data.dtypes == float)

    assert all(col in data.columns for col in expected_columns)

@pytest.mark.skipif(not sbr_webdriver, reason="Skipping Bitbo integration tests: SBR_WEBDRIVER not provided.")
@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://charts.bitbo.io/cycle-repeat/",
        ['MA1458d', 'MA200d', 'Price end of day']
    ),
])
def test_bitbo_download_brightdata(url, expected_columns):
    data = _download(url, sbr_webdriver=sbr_webdriver)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert all(data.dtypes == float)

    assert all(col in data.columns for col in expected_columns)



import pytest
import pandas as pd
from chaindl.scraper.dune import _download

def test_download_dune():
    url = "https://dune.com/queries/3265994/5466888"
    df = _download(url)
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "date" in df.columns
    assert "BTC_Price" in df.columns
    assert "mv_ratio" in df.columns

def test_download_dashboard():
    url = "https://dune.com/cryptokoryo/crypto-buy-signal"
    with pytest.raises(ValueError, match="URL is not a valid Dune query URL"):
        _download(url)



import pytest
import pandas as pd
from unittest.mock import patch, MagicMock
from chaindl.scraper.bitcoinmagazinepro import _download, _create_dataframes, _intercept_network_requests

@pytest.fixture
def mock_data():
    return {
        'response': {
            'chart': {
                'figure': {
                    'data': [
                        {
                            'name': 'Series1',
                            'x': ['2023-01-01', '2023-01-02', '2023-01-03'],
                            'y': [1, 2, 3],
                            'line': {}
                        },
                        {
                            'name': 'Series2',
                            'x': ['2023-01-01', '2023-01-02', '2023-01-03'],
                            'y': [4, 5, 6],
                            'line': {}
                        }
                    ]
                }
            }
        }
    }

def test_create_dataframes(mock_data):
    traces = mock_data['response']['chart']['figure']['data']
    dfs = _create_dataframes(traces)

    expected_df1 = pd.DataFrame({
        'Series1': [1, 2, 3]
    }, index=pd.to_datetime(['2023-01-01', '2023-01-02', '2023-01-03']))
    expected_df1.index.name = 'Date'

    expected_df2 = pd.DataFrame({
        'Series2': [4, 5, 6]
    }, index=pd.to_datetime(['2023-01-01', '2023-01-02', '2023-01-03']))
    expected_df2.index.name = 'Date'

    pd.testing.assert_frame_equal(dfs[0], expected_df1)
    pd.testing.assert_frame_equal(dfs[1], expected_df2)

@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://www.bitcoinmagazinepro.com/charts/sopr-spent-output-profit-ratio/",
        ['BTC Price', 'SOPR']
    ),
    (
        "https://www.bitcoinmagazinepro.com/charts/stock-to-flow-model/",
        ['BTC Price', 'Stock/Flow (365d average)', 'Model Variance', 'Halving Dates']
    )
])
def test_bitbo_download(url, expected_columns):
    data = _download(url)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert all(data.dtypes == float)

    assert all(col in data.columns for col in expected_columns)



import os
import pytest
import pandas as pd
from dotenv import load_dotenv

from chaindl.scraper.cryptoquant import _download, _create_dataframe

load_dotenv()

email = os.getenv('CRYPTOQUANT_EMAIL')
password = os.getenv('CRYPTOQUANT_PASSWORD')

@pytest.fixture
def sample_results():
    return [
        ['2023-01-01', 100],
        ['2023-01-02', 200],
        ['2023-01-03', 300]
    ]

@pytest.fixture
def sample_columns():
    return ['day', 'value']

def test_download_missing_credentials():
    with pytest.raises(TypeError, match="Email and/or password hasn't been passed"):
        _download('https://cryptoquant.com/some/path/123')

def test_download_unsupported_url():
    with pytest.raises(NotImplementedError, match="Only third party metrics on cryptoquant have been implemented."):
        _download('https://cryptoquant.com/asset/some/path/123', email='test@example.com', password='password123')

def test_create_dataframe(sample_results, sample_columns):
    expected_df = pd.DataFrame({
        'value': [100, 200, 300]
    }, index=pd.to_datetime(['2023-01-01', '2023-01-02', '2023-01-03']))
    expected_df.index.name = 'Date'

    df = _create_dataframe(sample_results, sample_columns)

    pd.testing.assert_frame_equal(df, expected_df)

@pytest.mark.skipif(not email or not password, reason="Skipping cryptoquant integration tests: Email and/or password not provided.")
@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://cryptoquant.com/analytics/query/66636764d376670a9fe4c3f4?v=66636764d376670a9fe4c3f6",
        ['avg_daily_price', 'avg_daily_realized_profit', 'avg_daily_realized_loss', 'sma_30d_realized_profit', 'sma_30d_realized_loss', 'ratio_30d_realized_profit_loss', 'Low_Liquidity']
    ),
    (
        "https://cryptoquant.com/analytics/query/6463b524885a7d37a1630f8b?v=6463f8c9fb92892124bd5864",
        ['Price', 'Adjusted_MVRV', 'mvrv', 'oversold', 'overbought']
    )
])
def test_cryptoquant(url, expected_columns):
    data = _download(url, email=email, password=password)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert all(data.dtypes == float)

    assert all(col in data.columns for col in expected_columns)



import time
import json
import pandas as pd
from seleniumwire import webdriver
from selenium.webdriver.chrome.options import Options
from seleniumwire.utils import decode

def _download(url, **kwargs):
    data = _intercept_network_requests(url, **kwargs)
    traces = data['response']['chart']['figure']['data']
    dfs = _create_dataframes(traces)
    merged_df = pd.concat(dfs, axis=1, join='outer')
    return merged_df

def _create_dataframes(traces):
    dfs = []
    for trace in traces:
        # if 'customdata' in trace:
            name = trace['name']
            x = trace['x']
            y = trace['y']

            length = min(len(x), len(y))
            x = x[:length]
            y = y[:length]

            df = pd.DataFrame({ name: pd.to_numeric(y, errors='coerce') }, index=pd.to_datetime(pd.to_datetime(x, format='mixed').date))
            df = df[~df.index.duplicated(keep='first')]
            df.index.name = 'Date'
            dfs.append(df)

    return dfs

def _intercept_network_requests(url, check_interval=0.5, timeout=30):
    # Set up Chrome options for headless mode
    chrome_options = Options()
    chrome_options.add_argument('--headless')  # Enable headless mode
    chrome_options.add_argument('--disable-gpu')  # Disable GPU for compatibility
    chrome_options.add_argument('--no-sandbox')  # Bypass OS security model

    driver = webdriver.Chrome(options=chrome_options)
    driver.get(url)

    start_time = time.time()
    request = None

    while time.time() - start_time < timeout:
        for req in driver.requests:
            if "_dash-update-component" in req.url and req.response:
                request = req
                break
        if request:
            break
        time.sleep(check_interval)

    if request:
        content_encoding = request.response.headers.get('Content-Encoding', '')
        body = decode(request.response.body, content_encoding)
        body = body.decode('utf-8', errors='ignore')
        driver.quit()
        return json.loads(body)
    else:
        driver.quit()
        raise TimeoutError(f"Could not find the request within {timeout} seconds. Try increasing the timeout!")



import re
import json
import time
import pandas as pd

from selenium.webdriver import Remote, ChromeOptions
from selenium.webdriver.chromium.remote_connection import ChromiumRemoteConnection
from selenium.webdriver.common.by import By

from seleniumbase import SB
from selenium.common.exceptions import StaleElementReferenceException

def _download(url, **kwargs):
    content = _get_script_content(url, **kwargs)
    traces = _get_traces(content)

    dfs = []
    for trace in traces:
        name, x, y = _get_data(trace, content)

        df = pd.DataFrame({ name: pd.to_numeric(y, errors='coerce') }, index=pd.to_datetime(x))
        df.index.name = 'Date'
        dfs.append(df)

    merged_df = pd.concat(dfs, axis=1, join='outer')
    return merged_df

def _get_script_content(url, **kwargs):
    sbr_webdriver = kwargs.get('sbr_webdriver')
    if sbr_webdriver:
        return _get_script_content_brightdata(url, sbr_webdriver)
    else:
        return _get_script_content_seleniumbase(url)

def _get_script_content_brightdata(url, sbr_webdriver):
    sbr_connection = ChromiumRemoteConnection(sbr_webdriver, 'goog', 'chrome')
    with Remote(sbr_connection, options=ChromeOptions()) as driver:
        driver.get(url)

        # CAPTCHA handling: If you're expecting a CAPTCHA on the target page, use the following code snippet to check the status of Scraping Browser's automatic CAPTCHA solver
        print('Waiting captcha to solve...')
        solve_res = driver.execute('executeCdpCommand', {
            'cmd': 'Captcha.waitForSolve',
            'params': {'detectTimeout': 20000},
        })
        print('Captcha solve status:', solve_res['value']['status'])

        script_content = ""
        script_tags = driver.find_elements(By.TAG_NAME, 'script')
        for script_tag in script_tags:
            script_inner_html = script_tag.get_attribute("innerHTML")
            if script_inner_html and 'trace' in script_inner_html:
                script_content += script_inner_html
    
    return script_content

def _get_script_content_seleniumbase(url):
    script_content = ""
    with SB(uc=True) as sb:
        sb.uc_open_with_reconnect(url, 4)
        sb.uc_gui_click_captcha()

        attempts = 0
        while attempts < 3:
            try:
                script_tags = sb.find_elements("script")
                for script_tag in script_tags:
                    script_inner_html = script_tag.get_attribute("innerHTML")
                    if script_inner_html and 'trace' in script_inner_html:
                        script_content += script_inner_html
                break
            except StaleElementReferenceException:
                attempts += 1
                time.sleep(1)

    return script_content

def _get_traces(content):
    trace_pattern = r'var\s+trace\d+\s*=\s*(\{.*?\});'
    traces = re.findall(trace_pattern, content, re.DOTALL)
    return traces

def _get_data(trace, content):
    x_pattern = r'x:\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*,'
    y_pattern = r'y:\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*,'
    name_pattern = r"name:\s*'([^']*)'"
    var_pattern = r'var\s+({name})\s*=\s*([^;]*);'

    name = ""
    x, y = [], []

    x_match = re.search(x_pattern, trace)
    y_match = re.search(y_pattern, trace)

    if x_match and y_match:
        x_var_name = x_match.group(1)
        y_var_name = y_match.group(1)

        x_var_pattern = var_pattern.format(name=x_var_name)
        y_var_pattern = var_pattern.format(name=y_var_name)

        x = re.search(x_var_pattern, content)
        y = re.search(y_var_pattern, content)

        if x and y:
            x = json.loads(x.group(2))
            y = json.loads(y.group(2))

            length = min(len(x), len(y))
            x = x[:length]
            y = y[:length]

            name_match = re.search(name_pattern, trace)
            if name_match:
                name = name_match.group(1)
    
    return name, x, y



import pytest
from chaindl.scraper.utils import _join_url

@pytest.mark.parametrize(
    "base_url, path, expected", [
        ("https://example.com/", "/path/to/resource", "https://example.com/path/to/resource"),
        ("https://www.example.com", "path/to/resource", "https://www.example.com/path/to/resource"),
        ("https://www.example.com", "/path/to/resource", "https://www.example.com/path/to/resource")
    ]
)
def test_join_url(base_url, path, expected):
    assert _join_url(base_url, path) == expected



from chaindl.scraper.theblock import _download

def test_transactions_on_the_bitcoin_network():
    url = "https://www.theblock.co/data/on-chain-metrics/bitcoin/transactions-on-the-bitcoin-network-daily"
    df = _download(url)
    assert "7DMA" in df.columns
    assert "Date" == df.index.name
    assert not df.empty



import pandas as pd
import pytest
from unittest.mock import patch
from chaindl import download

mocked_data_with_dates = pd.DataFrame({
    'data': [1, 2, 3, 4, 5]
}, index=pd.date_range(start='2023-01-01', periods=5))

mocked_data = {
    "checkonchain": mocked_data_with_dates,
    "chainexposed": mocked_data_with_dates,
    "bitbo": mocked_data_with_dates,
    "woocharts": mocked_data_with_dates,
    "cryptoquant": mocked_data_with_dates,
    "bitcoinmagazinepro": mocked_data_with_dates
}

@pytest.mark.parametrize("url, expected_data, provider", [
    ("https://charts.checkonchain.com/some_data", mocked_data["checkonchain"], "checkonchain"),
    ("https://chainexposed.com/some_data", mocked_data["chainexposed"], "chainexposed"),
    ("https://charts.bitbo.io/some_data", mocked_data["bitbo"], "bitbo"),
    ("https://woocharts.com/some_data", mocked_data["woocharts"], "woocharts"),
    ("https://cryptoquant.com/some_data", mocked_data["cryptoquant"], "cryptoquant"),
    ("https://www.bitcoinmagazinepro.com/some_data", mocked_data["bitcoinmagazinepro"], "bitcoinmagazinepro"),
])
def test_download_valid_urls(url, expected_data, provider):
    with patch(f'chaindl.scraper.{provider}._download', return_value=expected_data):
        result = download(url)
        pd.testing.assert_frame_equal(result, expected_data)
    
def test_download_invalid_url():
    invalid_url = "https://unknownsource.com/some_data"
    with pytest.raises(ValueError, match="Unsupported source"):
        download(invalid_url)


@pytest.mark.parametrize("url, start, end, expected_data", [
    ("https://charts.checkonchain.com/some_data", '2023-01-02', '2023-01-04', 
     pd.DataFrame({'data': [2, 3, 4]}, index=pd.date_range(start='2023-01-02', periods=3))),
    ("https://charts.checkonchain.com/some_data", '2023-01-01', None, 
     pd.DataFrame({'data': [1, 2, 3, 4, 5]}, index=pd.date_range(start='2023-01-01', periods=5))),
    ("https://charts.checkonchain.com/some_data", None, '2023-01-03', 
     pd.DataFrame({'data': [1, 2, 3]}, index=pd.date_range(start='2023-01-01', periods=3))),
])
def test_download_date_filtration(url, start, end, expected_data):
    with patch('chaindl.scraper.checkonchain._download', return_value=mocked_data["checkonchain"]):
        result = download(url, start=start, end=end)
        pd.testing.assert_frame_equal(result, expected_data)



import re
import json

import pandas as pd
from bs4 import BeautifulSoup

from . import utils

def _download(url):
    content = utils._get_page_content(url)
    soup = BeautifulSoup(content, 'html.parser')
    scripts = soup.find_all('script')

    dfs = _extract_data_from_scripts(scripts)

    merged_df = pd.concat(dfs, axis=1, join='outer')
    return merged_df

def _extract_data_from_scripts(scripts):
    dfs = []

    for script in scripts:
        if script.string and 'Plotly.newPlot' in script.string:
            matches = re.findall(r'var trace\d+ =\s*{\s*x:\s*(\[[^\]]*\]),\s*y:\s*(\[[^\]]*\]),.*?name:\s*\'(.*?)\'', script.string, re.DOTALL)
            for match in matches:
                x_data, y_data, name = match
                name = name.replace('\\u003c', '<').replace('\\u003e', '>')
                x = json.loads(x_data)
                y = json.loads(y_data)

                df = pd.DataFrame({ name: pd.to_numeric(y, errors='coerce') }, index=pd.to_datetime(pd.to_datetime(x).date))
                df.index.name = 'Date'
                dfs.append(df)

    return dfs



import requests
import pandas as pd

def _download(url, timespan="all", daysAverage="1d", include_price=True):
    """Allowed timespans: 30days, 90days, 180days, 1year, 3years, all"""
    metric = url.rstrip("/").split("/")[-1]
    api_url = f"https://api.blockchain.info/charts/{metric}"
    params = {
        "timespan": timespan,
        "sampled": "true",
        "metadata": "false",
        "daysAverageString": daysAverage,
        "format": "json"
    }
    response = requests.get(api_url, params=params)
    data = response.json()
    if 'values' not in data or 'name' not in data:
        raise ValueError(f"Invalid response for {metric} from Blockchain API.")
    values = data['values']
    name = data['name']

    df = pd.DataFrame(values)
    df = df.rename(columns={"x": "Date", "y": name or metric})
    df['Date'] = pd.to_datetime(df['Date'], unit='s')
    df = df.set_index("Date")

    if include_price and metric != "market-price":
        # Fetch bitcoin price
        btc_url = "https://api.blockchain.info/charts/market-price"
        btc_response = requests.get(btc_url, params=params)
        btc_data = btc_response.json()
        if 'values' not in btc_data or 'name' not in btc_data:
            raise ValueError("Invalid response for market price from Blockchain API.")
        btc_values = btc_data['values']
        btc_df = pd.DataFrame(btc_values)
        btc_df = btc_df.rename(columns={"x": "Date", "y": "Market Price (USD)"})
        btc_df['Date'] = pd.to_datetime(btc_df['Date'], unit='s')
        btc_df = btc_df.set_index("Date")

        # Merge the dataframes on the Date index
        df = df.merge(btc_df, left_index=True, right_index=True, how='outer')

    return df



from .checkonchain import _download
from .chainexposed import _download
from .bitbo import _download
from .woocharts import _download
from .cryptoquant import _download
from .bitcoinmagazinepro import _download
from .blockchain import _download
from .glassnode import _download
from .theblock import _download
from .dune import _download



import pytest
import pandas as pd
from chaindl.scraper.checkonchain import _download
from chaindl.scraper.checkonchain import _extract_data_from_scripts
from chaindl.scraper import utils

def test_extract_data_from_scripts():
    scripts = [
        type('', (object,), {"string": """Plotly.newPlot("plot", [{
            "name": "Price",
            "x": ["2024-09-30", "2024-10-01", "2024-10-02"],
            "y": [1, 2, 3]
        }]);"""})()
    ]

    expected_data = pd.DataFrame({
        "Price": [1, 2, 3]
    }, index=pd.to_datetime(["2024-09-30", "2024-10-01", "2024-10-02"]))
    expected_data.index.name = 'Date'

    dfs = _extract_data_from_scripts(scripts)

    pd.testing.assert_frame_equal(dfs[0], expected_data)

def test_checkonchain_mocked(monkeypatch):
    def mock_get_page_content(url):
        return """
        <script>
        Plotly.newPlot("plot", [{
            "name": "Price",
            "x": ["2024-09-30", "2024-10-01", "2024-10-02"],
            "y": [1, 2, 3]
        }]);
        </script>
        """

    monkeypatch.setattr(utils, "_get_page_content", mock_get_page_content)

    url = "https://charts.checkonchain.com/test"
    data = _download(url)

    expected_data = pd.DataFrame({
        "Price": [1, 2, 3]
    }, index=pd.to_datetime(["2024-09-30", "2024-10-01", "2024-10-02"]))
    expected_data.index.name = 'Date'

    pd.testing.assert_frame_equal(data, expected_data)

@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://charts.checkonchain.com/btconchain/pricing/pricing_mayermultiple_zscore/pricing_mayermultiple_zscore_light.html",
        ['Mayer Multiple Z', '200DMA', 'Price', '1.5sd', '1.0sd', '-1.0sd', '-1.5sd']
    ),
    (
        "https://charts.checkonchain.com/btconchain/realised/sopr/sopr_light.html",
        ['Price', 'Spent Output Profit Ratio (SOPR)', 'SOPR 7D-EMA']
    )
])
def test_real_checkonchain_download(url, expected_columns):
    data = _download(url)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert all(data.dtypes == float)

    assert all(col in data.columns for col in expected_columns)




import time
import json
from urllib.parse import urlparse
from selenium import webdriver
import pandas as pd

CRYPTOQUANT_URL = "https://cryptoquant.com/"

def _download(url, **kwargs):
    email = kwargs.get('email')
    password = kwargs.get('password')
    if not email or not password:
        raise TypeError("Email and/or password hasn't been passed")

    splits = urlparse(url).path.split('/')
    id = splits[-1]

    # Cryptoquant's own metrics
    if splits[1] == 'asset':
        raise NotImplementedError("Only third party metrics on cryptoquant have been implemented.")

    proxy = kwargs.get('proxy', None)
    driver = _get_driver(proxy=proxy)

    data = _get_json(driver, id, email, password)

    columns = data['data']['result']['columns']
    results = data['data']['result']['results']
    column_names = [col['name'] for col in columns]

    return _create_dataframe(results, column_names)

def _create_dataframe(results, column_names):
    df = pd.DataFrame(results, columns=column_names)
    
    date_column = None
    for col in df.columns:
        if col.lower() in ['day', 'date', 'datetime', 'transaction_day']:
            date_column = col
            break
    
    if date_column:
        df[date_column] = pd.to_datetime(df[date_column])
        df.set_index(date_column, inplace=True)
        df.index.name = 'Date'
    else:
        print("Unable to find and parse the date column")
    
    return df

def _get_driver(proxy=None):
    chrome_options = webdriver.ChromeOptions()

    if proxy:
        driver = webdriver.Remote(proxy, options=chrome_options)
    else:
        driver = webdriver.Chrome(options=chrome_options)
    
    return driver

def _get_json(driver, id, email, password):
    driver.get(CRYPTOQUANT_URL)
    time.sleep(4)

    # Execute the login request
    script = f"""
        return fetch("https://api.cryptoquant.com/live/v1/sign-in", {{
            method: 'POST',
            headers: {{
                'Content-Type': 'application/json'
            }},
            body: JSON.stringify({{
                "email": "{email}",
                "password": "{password}"
            }})
        }}).then(response => response.json());
    """
    response = driver.execute_script(script)

    if 'accessToken' in response:
        access_token = response['accessToken']

        data_url = f"https://api.cryptoquant.com/live/v1/analytics/{id}"
        
        result_script = f"""
            return fetch("{data_url}", {{
                method: 'GET',
                headers: {{
                    'Authorization': 'Bearer {access_token}',
                    'Accept': 'application/json'
                }}
            }}).then(response => response.json());
        """
        result = driver.execute_script(result_script)
    else:
        print(f"Error occurred: {response.get('error')}")
        driver.quit()
        return {}

    driver.quit()
    return result



import requests
import pandas as pd

from .utils import _join_url

def _download(url):
    data = _request_chart_json(url)
    
    dfs = []
    for key, value in data.items():
        name = key
        x = value['x']
        y = value['y']

        df = pd.DataFrame({ name: pd.to_numeric(y, errors='coerce') }, index=pd.to_datetime(x, unit='ms'))
        df.index.name = 'Date'
        dfs.append(df)
    
    merged_df = pd.concat(dfs, axis=1, join='outer')
    return merged_df

def _request_chart_json(url):
    json_url = _join_url(url, "data/chart.json")
    response = requests.get(json_url)
    return response.json()



import pandas as pd
import cloudscraper
from urllib.parse import urlparse

def _download(url: str) -> pd.DataFrame:
    # Parse the query ID from URL
    parsed = urlparse(url)
    parts = [part for part in parsed.path.split("/") if part]
    if len(parts) < 2 or parts[0] != "queries":
        raise ValueError("URL is not a valid Dune query URL. Follow the guide at: https://chaindl.readthedocs.io/#dune-dune-com")
    query_id = int(parts[1])

    scraper = cloudscraper.create_scraper()

    # Get latest result set ID
    graphql_url = "https://dune.com/public/graphql"
    payload = {
        "operationName": "GetLatestResultSetIds",
        "variables": {"queryId": query_id, "parameters": [], "canRefresh": True},
        "query": """
                query GetLatestResultSetIds($canRefresh: Boolean!, $queryId: Int!, $parameters: [ExecutionParameterInput!]) {
                  resultSetForQuery(canRefresh: $canRefresh, queryId: $queryId, parameters: $parameters) {
                    completedExecutionId
                    failedExecutionId
                    pendingExecutionId
                    __typename
                  }
                }
            """
    }

    response = scraper.post(graphql_url, json=payload)
    if response.status_code != 200:
        raise ConnectionError(f"Failed to get result set ID: {response.status_code}")
    json_data = response.json()
    execution_id = json_data["data"]["resultSetForQuery"]["completedExecutionId"]
    if execution_id is None:
        raise ValueError("No completed execution found for this query.")

    # Fetch all execution data with pagination
    all_data = []
    offset = 0
    limit = 9999999

    execution_url = "https://core-api.dune.com/public/execution"
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:142.0) Gecko/20100101 Firefox/142.0",
        "Accept": "*/*",
        "Accept-Language": "en-US,en;q=0.5",
        "Content-Type": "application/json",
        "Origin": "https://dune.com",
        "Referer": "https://dune.com/",
        "Connection": "keep-alive",
        "Pragma": "no-cache",
        "Cache-Control": "no-cache",
    }

    while True:
        payload = {
            "execution_id": execution_id,
            "query_id": query_id,
            "parameters": [],
            "pagination": {"limit": limit, "offset": offset}
        }
        response = scraper.post(execution_url, headers=headers, json=payload)
        if response.status_code != 200:
            raise ConnectionError(f"Failed to fetch execution data: {response.status_code}")
        json_data = response.json()
        execution_result = json_data.get('execution_succeeded')
        if not execution_result:
            raise ValueError("Execution failed or no data returned.")

        data = execution_result['data']
        total_row_count = execution_result['total_row_count']
        all_data.extend(data)

        if len(all_data) >= total_row_count:
            break
        offset = len(all_data)

    # Convert to DataFrame
    df = pd.DataFrame(all_data)
    return df



import json
import requests
import pandas as pd
from urllib.parse import urlparse

def _download(url):
    parsed = urlparse(url)
    parts = [p for p in parsed.path.split("/") if p]
    last_three = "/".join(parts[-3:])

    api = f"https://www.theblock.co/api/charts/chart/{last_three}"
    response = requests.get(api)
    response.raise_for_status()
    _json = response.json()
    jsonFile = _json.get("jsonFile", {}).get("data", {})
    jsonFile = json.loads(jsonFile)

    df = pd.concat(
        {
            series: pd.DataFrame(data["Data"]).set_index("Timestamp")["Result"]
            for series, data in jsonFile["Series"].items()
        },
        axis=1
    )
    df.index = pd.to_datetime(df.index, unit="s")
    df.index.name = "Date"

    return df



import pandas as pd

from . import scraper

def download(url, start=None, end=None, **kwargs):
    """
    Downloads cryptocurrency data from the specified URL and returns it as a pandas DataFrame.

    This function supports various data sources and handles specific URLs to retrieve data from each.

    Args:
        url (str): The URL from which to download the data. It must match one of the known data sources.
        start (str, optional): The start date for slicing the DataFrame. Must be in a format recognized by pandas (e.g., 'YYYY-MM-DD').
        end (str, optional): The end date for slicing the DataFrame. Must be in a format recognized by pandas (e.g., 'YYYY-MM-DD').
        **kwargs: Additional keyword arguments to pass to specific scraper methods.\n
            email and password needs to be passed for Cryptoquant\n
            sbr_webdriver needs to be passed for using a remote browser proxy (Eg: BrightData, etc.)

    Returns:
        pd.DataFrame: A DataFrame containing the downloaded data. The DataFrame index is datetime.

    Raises:
        ValueError: If the provided URL does not match any known data sources.

    Supported Data Sources:
        - CheckOnChain: "https://charts.checkonchain.com"
        - ChainExposed: "https://chainexposed.com"
        - BitBo: "https://charts.bitbo.io"
        - WooCharts: "https://woocharts.com"
        - CryptoQuant: "https://cryptoquant.com"
        - Bitcoin Magazine Pro: "https://www.bitcoinmagazinepro.com"
        - Blockchain.com: "https://www.blockchain.com/explorer/charts"
        - Glassnode: "https://studio.glassnode.com/charts"
        - The Block: "https://www.theblock.co"
        - Dune: "https://dune.com"

    Example:
        >>> df = download("https://charts.checkonchain.com/path/to/indicator")
        >>> df_filtered = download("https://charts.checkonchain.com/path/to/indicator", start='2023-01-01', end='2023-12-31')
        >>> cryptoquant = download("https://cryptoquant.com/path/to/indicator", email=email, password=password)
    """
    CHECKONCHAIN_BASE_URL = "https://charts.checkonchain.com"
    CHAINEXPOSED_BASE_URL = "https://chainexposed.com"
    BITBO_BASE_URL = "https://charts.bitbo.io"
    WOOCHARTS_BASE_URL = "https://woocharts.com"
    CRYPTOQUANT_BASE_URL = "https://cryptoquant.com"
    BITCOINMAGAZINEPRO_BASE_URL = "https://www.bitcoinmagazinepro.com"
    BLOCKCHAIN_BASE_URL = "https://www.blockchain.com/explorer/charts"
    GLASSNODE_BASE_URL = "https://studio.glassnode.com/charts"
    THEBLOCK_BASE_URL = "https://www.theblock.co"
    DUNE_BASE_URL = "https://dune.com"

    data = pd.DataFrame()

    if url.startswith(CHECKONCHAIN_BASE_URL):
        data = scraper.checkonchain._download(url)
    elif url.startswith(CHAINEXPOSED_BASE_URL):
        data = scraper.chainexposed._download(url)
    elif url.startswith(BITBO_BASE_URL):
        data = scraper.bitbo._download(url, **kwargs)
    elif url.startswith(WOOCHARTS_BASE_URL):
        data = scraper.woocharts._download(url)
    elif url.startswith(CRYPTOQUANT_BASE_URL):
        data = scraper.cryptoquant._download(url, **kwargs)
    elif url.startswith(BITCOINMAGAZINEPRO_BASE_URL):
        data = scraper.bitcoinmagazinepro._download(url, **kwargs)
    elif url.startswith(BLOCKCHAIN_BASE_URL):
        data = scraper.blockchain._download(url, **kwargs)
    elif url.startswith(GLASSNODE_BASE_URL):
        data = scraper.glassnode._download(url, **kwargs)
    elif url.startswith(THEBLOCK_BASE_URL):
        data = scraper.theblock._download(url)
    elif url.startswith(DUNE_BASE_URL):
        data = scraper.dune._download(url)
    else:
        raise ValueError("Unsupported source. Find the list of supported websites here: https://chaindl.readthedocs.io/")
    
    if pd.api.types.is_datetime64_any_dtype(data.index):
        if start:
            data = data.loc[start:]
        if end:
            data = data.loc[:end]

    return data



import pytest
import pandas as pd
from unittest.mock import Mock

from chaindl.scraper.woocharts import _request_chart_json, _download

@pytest.fixture
def mock_response():
    return {
        "chart1": {
            "x": [1620000000000, 1620003600000],
            "y": ["1.1", "1.2"]
        },
        "chart2": {
            "x": [1620000000000, 1620003600000],
            "y": ["2.1", "2.2"]
        }
    }

def test_request_chart_json(mocker, mock_response):
    url = "https://example.com"
    mocker.patch('requests.get') 
    mock_requests = mocker.patch('requests.get', return_value=Mock(json=lambda: mock_response))

    json_data = _request_chart_json(url)
    assert json_data == mock_response
    mock_requests.assert_called_once_with('https://example.com/data/chart.json')

def test_download(mocker, mock_response):
    url = "https://example.com"
    mocker.patch('requests.get', return_value=Mock(json=lambda: mock_response))

    expected_index = pd.to_datetime([1620000000000, 1620003600000], unit='ms')
    expected_df = pd.DataFrame({
        "chart1": pd.to_numeric([1.1, 1.2], errors='coerce'),
        "chart2": pd.to_numeric([2.1, 2.2], errors='coerce'),
    }, index=expected_index)
    expected_df.index.name = 'Date'
    
    result_df = _download(url)

    pd.testing.assert_frame_equal(result_df, expected_df)


@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://woocharts.com/bitcoin-price-models/",
        ['market', 'realised', 'delta', 'top_', 'vwap', 'wma200', 'cvdd', 'balanced']
    ),
    (
        "https://woocharts.com/bitcoin-congestion/",
        ['volume', 'size_block', 'size_total', 'confirm_time', 'adoption', 'mempool']
    )
])
def test_woocharts(url, expected_columns):
    data = _download(url)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert all(data.dtypes == float)

    assert all(col in data.columns for col in expected_columns)



import pandas as pd
import pytest

from chaindl.scraper.blockchain import _download

@pytest.mark.parametrize("url, expected_columns", [
    (
        "https://www.blockchain.com/explorer/charts/avg-block-size",
        ['Market Price (USD)', 'Average Block Size']
    ),
    (
        "https://www.blockchain.com/explorer/charts/market-price",
        ['Market Price (USD)']
    )
])
def test_blockchain_download(url, expected_columns):
    data = _download(url)

    assert isinstance(data, pd.DataFrame)
    assert isinstance(data.index, pd.DatetimeIndex)

    print(data)
    assert all(col in data.columns for col in expected_columns)



import re
from urllib.parse import urlparse, parse_qs
import requests
import pandas as pd

SESSION = requests.Session()

def _parse_metric_path(url: str):
    """Convert Glassnode chart URL path into API metric path and asset."""
    o = urlparse(url)
    path = o.path.replace(".", "/")
    query_params = parse_qs(o.query)
    asset = query_params.get("a", ["BTC"])[0]

    try:
        _, charts, *prefix_parts, last_segment = path.split("/")
    except ValueError:
        raise ValueError(f"Invalid Glassnode URL path: {path}")

    # Step 1: handle letter number boundaries: Min1Count -> Min_1Count
    number_adjusted = re.sub(r'([a-zA-Z])([0-9])', r'\1_\2', last_segment)
    # Step 2: handle Min_1Count -> Min_1_Count
    snake_case = re.sub(r'(?<!^)(?=[A-Z])', '_', number_adjusted).lower()
    new_path = "/".join(prefix_parts + [snake_case])
    return new_path, asset, snake_case

def _fetch_json(url: str):
    """Fetch JSON from a URL with session handling and error checking."""
    resp = SESSION.get(url)
    resp.raise_for_status()
    try:
        return resp.json()
    except ValueError:
        raise ValueError(f"Failed to parse JSON from {url}: {resp.text}")

def _process_metric_json(json_data, column_name: str):
    """Convert Glassnode metric JSON into a clean DataFrame."""
    df = pd.DataFrame(json_data)

    if 'o' in df.columns:
        # Expand nested object
        columns = list(df['o'].iloc[0].keys())
        expanded_df = pd.DataFrame([row['o'] for row in json_data])
        expanded_df['Date'] = df['t']
        cols = ['Date'] + columns
        expanded_df = expanded_df[cols]
        expanded_df.index = pd.to_datetime(expanded_df['Date'], unit='s')
        expanded_df = expanded_df.drop(columns=['Date'])
        return expanded_df
    else:
        df.rename(columns={'t': 'Date', 'v': column_name}, inplace=True)
        df.index = pd.to_datetime(df['Date'], unit='s')
        df = df.drop(columns=['Date'])
        return df

def _download(url: str, include_price: bool = True) -> pd.DataFrame:
    try:
        metric_path, asset, snake_case = _parse_metric_path(url)

        # Fetch cookies
        SESSION.get(url)

        # Fetch metric
        api_url = f"https://api.glassnode.com/v1/metrics/{metric_path}?a={asset}&referrer=charts"
        json_data = _fetch_json(api_url)

        if 'requiredPlan' in json_data:
            raise ValueError(f"Metric {snake_case} is not available on the free plan.")

        df = _process_metric_json(json_data, snake_case)

        # Optionally fetch USD price
        if include_price:
            price_api = f"https://api.glassnode.com/v1/metrics/market/price_usd_close?a={asset}&referrer=charts"
            json_price = _fetch_json(price_api)
            df_price = pd.DataFrame(json_price)
            df_price.rename(columns={'t': 'Date', 'v': 'price_usd_close'}, inplace=True)
            df_price.index = pd.to_datetime(df_price['Date'], unit='s')
            df_price = df_price.drop(columns=['Date'])
            df = df.join(df_price, how='outer')

        return df

    except requests.exceptions.RequestException as e:
        raise RuntimeError(f"Network error while fetching Glassnode metric: {e}")
    except ValueError as e:
        raise RuntimeError(f"Data error: {e}")
    except Exception as e:
        raise RuntimeError(f"Unexpected error: {e}")



import pytest
import pandas as pd
from unittest.mock import patch, Mock
from chaindl.scraper import glassnode

def test_parse_metric_path_basic():
    url = "https://studio.glassnode.com/charts/addresses.ActiveCountWithContracts?a=ETH"
    path, asset, snake_case = glassnode._parse_metric_path(url)
    assert asset == "ETH"
    assert "active_count_with_contracts" in snake_case
    assert path.endswith(snake_case)

def test_parse_metric_path_default_asset():
    url = "https://studio.glassnode.com/charts/addresses.ActiveCountWithContracts"
    path, asset, snake_case = glassnode._parse_metric_path(url)
    assert asset == "BTC"


def test_process_metric_json_simple():
    json_data = [{"t": 1680000000, "v": 42}, {"t": 1680000600, "v": 43}]
    df = glassnode._process_metric_json(json_data, "metric_name")
    assert isinstance(df, pd.DataFrame)
    assert "metric_name" in df.columns
    assert df.index.name is 'Date'

def test_process_metric_json_nested():
    json_data = [{"t": 1680000000, "o": {"a": 1, "b": 2}}, {"t": 1680000600, "o": {"a": 3, "b": 4}}]
    df = glassnode._process_metric_json(json_data, "ignored")
    assert isinstance(df, pd.DataFrame)
    assert "a" in df.columns and "b" in df.columns
    assert df.shape == (2, 2)  # Only a and b remain

@patch("chaindl.scraper.glassnode.SESSION.get")
def test_fetch_json_success(mock_get):
    mock_resp = Mock()
    mock_resp.raise_for_status.return_value = None
    mock_resp.json.return_value = {"key": "value"}
    mock_get.return_value = mock_resp
    data = glassnode._fetch_json("https://fakeurl.com")
    assert data["key"] == "value"

@patch("chaindl.scraper.glassnode.SESSION.get")
def test_download_includes_price(mock_get):
    # Mock metric response
    metric_resp = Mock()
    metric_resp.raise_for_status.return_value = None
    metric_resp.json.return_value = [{"t": 1680000000, "v": 100}]
    # Mock price response
    price_resp = Mock()
    price_resp.raise_for_status.return_value = None
    price_resp.json.return_value = [{"t": 1680000000, "v": 50000}]

    mock_get.side_effect = [Mock(), metric_resp, price_resp]  # First GET for cookies
    url = "https://studio.glassnode.com/charts/addresses.ActiveCountWithContracts?a=ETH"

    df = glassnode._download(url)
    assert "price_usd_close" in df.columns
    assert "active_count_with_contracts" in df.columns
    assert isinstance(df, pd.DataFrame)

@patch("chaindl.scraper.glassnode.SESSION.get")
def test_download_raises_on_required_plan(mock_get):
    resp = Mock()
    resp.raise_for_status.return_value = None
    resp.json.return_value = {"requiredPlan": True}
    mock_get.side_effect = [Mock(), resp]

    url = "https://studio.glassnode.com/charts/addresses.ActiveCountWithContracts?a=ETH"
    with pytest.raises(RuntimeError, match="Data error"):
        glassnode._download(url)

# Integration tests
def test_download_active_count_btc():
    url = "https://studio.glassnode.com/charts/addresses.ActiveCount?a=BTC"
    df = glassnode._download(url)
    assert "active_count" in df.columns
    assert "price_usd_close" in df.columns
    assert not df.empty

def test_download_stock_to_flow_btc():
    url = "https://studio.glassnode.com/charts/indicators.StockToFlowRatio?a=BTC"
    df = glassnode._download(url)
    assert "daysTillHalving" in df.columns
    assert "price" in df.columns
    assert "ratio" in df.columns
    assert "price_usd_close" in df.columns
    assert not df.empty

def test_download_invalid_url():
    url = "https://studio.glassnode.com/charts/addresses.NotARealMetric?a=BTC"
    with pytest.raises(RuntimeError):
        glassnode._download(url)


import requests

def _get_page_content(url):
    response = requests.get(url)
    response.raise_for_status()
    return response.text

def _join_url(base_url, path):
    return base_url.rstrip('/') + '/' + path.lstrip('/')
